from abc import ABC, abstractmethod
from typing import Literal

from mako.template import Template

_DEFAULT_PROMPT = Template(
    """
Given a python code, that could contains class(es), function(s), method(s), variable(s), etc. generate a docstring for it using the ${docstring_format} format.

% if not (skip_raises and skip_returns and skip_examples):
The generated docstring should include for each class, function, method, variable, etc.:
% if not skip_raises:
- The exceptions that the code can raise.
% endif
% if not skip_returns:
- The return value of the code.
% endif
% if not skip_examples:
- Examples of how to use the code.
% endif

% endif

The output code must be only the input code populated with the generated docstrings.
No additional code or comments should be added.
No bug fixing or code refactoring should be done.

Return the output code in the following format:



```generated-python-code
<OUTPUT_CODE>
```

No other output format is accepted.
"""
)


class Genius(ABC):
    """
    An abstract class for generating docstrings for Python code.

    Args:
        docstring_format (Literal["google", "numpy", "sprinx"], optional): The format of the docstring. Defaults to "google".
        skip_raises (bool, optional): Whether to include exceptions in the docstring. Defaults to False.
        skip_returns (bool, optional): Whether to include return values in the docstring. Defaults to False.
        skip_examples (bool, optional): Whether to include examples in the docstring. Defaults to False.

    Raises:
        ValueError: If the provided docstring_format is not valid.
    """

    def __init__(
        self,
        docstring_format: Literal["google", "numpy", "sprinx"] = "google",
        skip_raises: bool = False,
        skip_returns: bool = False,
        skip_examples: bool = False,
        **kwargs,
    ):
        self.docstring_format = docstring_format
        if docstring_format not in ["google", "numpy", "sprinx"]:
            raise ValueError(
                f"Invalid docstring format {docstring_format}. Must be one of 'google', 'numpy', or 'sprinx'."
            )
        self.skip_raises = skip_raises
        self.skip_returns = skip_returns
        self.skip_examples = skip_examples
        super().__init__()

    @property
    def _prompt(self) -> Template:
        """
        Return the default prompt for the client.
        """
        return _DEFAULT_PROMPT

    @property
    def _hydrated_prompt(self) -> str:
        """
        Return the prompt with the necessary variables replaced with their values.
        """
        return self._prompt.render(
            docstring_format=self.docstring_format,
            skip_raises=self.skip_raises,
            skip_returns=self.skip_returns,
            skip_examples=self.skip_examples,
        )

    @abstractmethod
    def _generate_docstring(self, code: str) -> str:
        """
        Generate a docstring for the given code and return the code with the docstring added.

        Parameters:
            code (str): The code to generate a docstring for.

        Returns:
            str: The code with the generated docstring added.
        """
        pass

    def clean_llm_output(self, output: str) -> str:
        """
        Clean the output from the LLM model by removing any leading or trailing whitespace.

        Parameters:
            output (str): The output from the LLM model.

        Returns:
            str: The cleaned output.
        """
        import re

        output = re.findall(r"```generated-python-code\n(.*?)\n```", output, re.DOTALL)
        if output:
            return output[0].strip()
        return output

    def generate(self, code: str) -> str:
        """
        Generate a docstring for the given code.

        Parameters:
            code (str): The code to generate a docstring for.

        Returns:
            str: The code with the generated docstring added.
        """
        if not code:
            return code
        output = self._generate_docstring(code)
        return self.clean_llm_output(output)



from docugenius.genius import openai_genius
from docugenius.genius.genius import Genius

SUPPORTED_MODELS = ["openai:gpt-4o", "openai:gpt-4o-mini"]


def genius_factory(model: str, **kwargs) -> Genius:
    if model not in SUPPORTED_MODELS:
        raise ValueError(f"Unsupported model {model}. Must be one of {SUPPORTED_MODELS}.")

    if model in ["openai:gpt-4o", "openai:gpt-4o-mini"]:
        model = model.split(":")[1]
        return openai_genius.OpenAIGenius(model=model, **kwargs)



import os
from pathlib import Path
from typing import List


def find_python_files(root_dir: str) -> List[str]:
    """
    Return a list of all the Python files in the specified directory and its subdirectories.

    Args:
        root_dir (str): The root directory to search for Python files.

    Returns:
        List[str]: A list of all the Python files in the specified directory and its subdirectories.

    Raises:
        FileNotFoundError: If the specified root directory does not exist.
        PermissionError: If there is a permission error accessing the directory.

    Examples:
        >>> find_python_files('/path/to/directory')
        ['/path/to/directory/file1.py', '/path/to/directory/subdir/file2.py']
    """

    file_list = []
    root_path = Path(root_dir)
    for root, _, files in root_path.walk():
        for file in files:
            if file.endswith(".py"):
                file_list.append(os.path.join(root, file))
    return file_list



from typing import Literal

from docugenius.genius.genius import Genius


class OpenAIGenius(Genius):
    """
    A class to generate docstrings using OpenAI's API.

    Args:
        model (str): The model to use for generating docstrings.
        docstring_format (Literal["google", "numpy", "sprinx"], optional): The format of the docstring. Defaults to "google".
        skip_raises (bool, optional): Whether to include exceptions in the docstring. Defaults to False.
        skip_returns (bool, optional): Whether to include return values in the docstring. Defaults to False.
        skip_examples (bool, optional): Whether to include examples in the docstring. Defaults to False.

    Raises:
        ValueError: If the provided model is not valid.

    Examples:
        >>> genius = OpenAIGenius(model="gpt-4o-mini")
        >>> print(genius.docstring_format)
        google
    """

    def __init__(
        self,
        model: str,
        docstring_format: Literal["google", "numpy", "sprinx"] = "google",
        skip_raises: bool = False,
        skip_returns: bool = False,
        skip_examples: bool = False,
    ):
        """
        Initialize the OpenAIGenius instance.

        Args:
            model (str): The model to use for generating docstrings.
            docstring_format (Literal["google", "numpy", "sprinx"], optional): The format of the docstring. Defaults to "google".
            skip_raises (bool, optional): Whether to include exceptions in the docstring. Defaults to False.
            skip_returns (bool, optional): Whether to include return values in the docstring. Defaults to False.
            skip_examples (bool, optional): Whether to include examples in the docstring. Defaults to False.

        Raises:
            ValueError: If the provided model is not valid.

        Examples:
            >>> genius = OpenAIGenius(model="gpt-4o-mini")
        """
        import openai

        super().__init__(docstring_format, skip_raises, skip_returns, skip_examples)
        self._client = openai.OpenAI()
        self.model = model

    def _generate_docstring(self, code: str) -> str:
        """
        Generate a docstring for the provided code using OpenAI's API.

        Args:
            code (str): The code for which to generate a docstring.

        Returns:
            str: The generated docstring.

        Raises:
            Exception: If the OpenAI API call fails.

        Examples:
            >>> docstring = genius._generate_docstring("def example_function(): pass")
            >>> print(docstring)
        """
        response = self._client.chat.completions.create(
            model=self.model,
            messages=[
                {"role": "system", "content": self._hydrated_prompt},
                {"role": "user", "content": code},
            ],
            temperature=0,
            top_p=1,
        )
        return response.choices[0].message.content



from argparse import ArgumentParser
from pathlib import Path
from typing import Literal

from pydantic import BaseModel

from docugenius.common.file_utils import find_python_files
from docugenius.genius import genius_factory


class RunArgs(BaseModel):
    """
    A class to hold the arguments for running the docstring generation.

    Attributes:
        input_path (str): The path to the input file or directory.
        model (str): The model to use for generating docstrings.
                     Must be one of ["openai:gpt-4o", "openai:gpt-4o-mini"].
        docstring_format (str): The format of the generated docstrings.
                                Must be one of ["google", "numpy", "sprinx"].
        skip_raises (bool): Whether to include information about exceptions raised by the code.
        skip_returns (bool): Whether to include information about the return value of the code.
        skip_examples (bool): Whether to include examples of how to use the code.
        output_path (str): The path to the output file/destination.
                           If not provided, the output will overwrite the input file.
                           If a directory is passed as input, this should be a directory.
    Raises:
        ValueError: If the model or docstring_format is not one of the specified literals.
    """

    input_path: str
    model: str = Literal["openai:gpt-4o", "openai:gpt-4o-mini"]
    docstring_format: Literal["google", "numpy", "sprinx"] = "google"
    skip_raises: bool = True
    skip_returns: bool = True
    skip_examples: bool = True
    output_path: str | None = None

    def is_recursive(self) -> bool:
        """
        Check if the input path is a directory or a file.

        Returns:
            bool: True if the input path is a directory, False if it is a file.
        """
        return Path(self.input_path).is_dir()


def run_on_file(input_file: str, model: str, output_file: str = None):
    """
    Generate docstrings for the code in the specified input file.

    Args:
        input_file (str): The path to the input file.
        model (str): The model to use for generating docstrings.
        output_file (str, optional): The path to the output file. Defaults to None.

    Raises:
        FileNotFoundError: If the input file does not exist.
        Exception: If there is an error during docstring generation.

    Returns:
        None: This function does not return a value.

    Example:
        >>> run_on_file('example.py', 'openai:gpt-4o')
    """

    with open(input_file) as f:
        code = f.read()

    genius = genius_factory(model)

    docstring = genius.generate(code)

    output_file = output_file or input_file

    with open(output_file, "w") as f:
        f.write(docstring)


def run(args: RunArgs):
    """
    Run the docstring generation process based on the provided arguments.

    Args:
        args (RunArgs): The arguments for running the docstring generation.

    Raises:
        FileNotFoundError: If the input file/directory does not exist.
        Exception: If there is an error during docstring generation.

    Returns:
        None: This function does not return a value.

    Example:
        >>> run(RunArgs(input_path='example.py', model='openai:gpt-4o'))
    """

    input_path = args.input_path
    model = args.model
    output_path = args.output_path

    if args.is_recursive():
        for file in find_python_files(input_path):
            run_on_file(file, model, output_path)
    else:
        run_on_file(input_path, model, output_path)


def main():
    parser = ArgumentParser()
    parser.add_argument(
        "input_path",
        help="The path to the input. You can also pass a directory to process all files in it.",
    )

    parser.add_argument(
        "--model",
        "-m",
        default="openai:gpt-4o",
        choices=["openai:gpt-4o", "openai:gpt-4o-mini"],
        help="The model to use for generating docstrings.",
    )

    parser.add_argument(
        "--docstring-format",
        "-d",
        default="google",
        choices=["google", "numpy", "sprinx"],
        help="The format of the generated docstrings.",
    )

    parser.add_argument(
        "--skip-raises",
        "-r",
        action="store_false",
        help="Whether to include information about exceptions raised by the code.",
    )

    parser.add_argument(
        "--skip-returns",
        "-R",
        action="store_false",
        help="Whether to include information about the return value of the code.",
    )

    parser.add_argument(
        "--skip-examples",
        "-e",
        action="store_false",
        help="Whether to include examples of how to use the code.",
    )
    parser.add_argument(
        "--output-path",
        "-o",
        default=None,
        help="The path to the output file/destination. If not provided, the output will overwrite the input file. If a directory is passed as input, this should be a directory.",
    )

    args = parser.parse_args()
    run(RunArgs(**vars(args)))


if __name__ == "__main__":
    main()
